/* *****************************************************************************
 * type/cas.c -- get the CAS type and data types out of raw identification data.
 * Copyright (C) 2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libg1m.
 * libg1m is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libg1m is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libg1m; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libg1m/internals.h>

/* ************************************************************************** */
/*  Local types                                                               */
/* ************************************************************************** */
/* Correspondance type */
struct type_corresp {
	/* identification */
	const char *datatype;

	/* libg1m MCS file type, flags */
	g1m_mcstype_t type;
	unsigned int flags;
};

/* ************************************************************************** */
/*  Correspondances                                                           */
/* ************************************************************************** */
/* All correspondances found by Tom Wheeley and Tom Lynn. Remarks:
 * - Correspondances with a NULL data type means the data type isn't to be
 *   read. */

#define TTERM {NULL, 0, 0}
static struct type_corresp cas_groups[] = {
	/* implemented */
	{"LT", g1m_mcstype_list, 0}, // list
	{"MT", g1m_mcstype_matrix, 0}, // matrix
	{"P1", g1m_mcstype_program, 0}, // one program
	{"PG", g1m_mcstype_program, 0}, // same than above
	{"PZ", g1m_mcstype_program, g1m_mcsflag_multiple}, // set of programs
	{"VM", g1m_mcstype_variable, 0}, // one variable

	/* not implemented yet */
	{"AA", 0, 0}, // dynamic graph functions
	{"AD", 0, 0}, // variable memory
	{"AL", 0, 0}, // all
	{"AM", 0, 0}, // alpha variable memory
	{"BU", 0, 0}, // backup
	{"DM", 0, 0}, // defined memory
	{"DC", 0, 0}, // color screenshot
	{"DD", 0, 0}, // monochrome screenshot
	{"EN", 0, 0}, // one editor file
	{"FN", 0, 0}, // set of editor files
	{"FT", 0, 0}, // ??? (cafix)
	{"F1", 0, 0}, // one function memory
	{"F6", 0, 0}, // set of function memories
	{"GA", 0, 0}, // set of graph functions
	{"GF", 0, 0}, // graph zoom factor (graph function?)
	{"GM", 0, 0}, // gmem (cafix)
	{"GR", 0, 0}, // graph range
	{"GT", 0, 0}, // function table
	{"MA", 0, 0}, // set of matrices
	{"PC", 0, 0}, // picture from 9xxx (cafix)
	{"PD", 0, 0}, // polynomial equations
	{"RF", 0, 0}, // ??? (cafix)
	{"RR", 0, 0}, // ??? (cafix)
	{"RT", 0, 0}, // recursion table
	{"SD", 0, 0}, // simultaneous equations
	{"SE", 0, 0}, // equation (cafix)
	{"SR", 0, 0}, // linear regression data
	{"SS", 0, 0}, // standard deviation data
	{"TR", 0, 0}, // ??? (cafix)
	{"WD", 0, 0}, // window data (cafix)

	/* terminating entry */
	TTERM
};

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/**
 *	g1m_maketype_cas:
 *	Get the libg1m MCS type from raw CAS identification data.
 *
 *	@arg	head		the head to fill.
 *	@arg	datatype	the data type string.
 *	@return				the error (if any).
 */

int g1m_maketype_cas(g1m_mcshead_t *head, const char *datatype)
{
	/* copy information */
	memset(head, 0, sizeof(g1m_mcshead_t));
	head->flags |= g1m_mcsinfo_cas;
	memcpy(head->_datatype, datatype, 2);
	head->_datatype[2] = 0;

	/* look for correspondance */
	struct type_corresp *c = cas_groups - 1;
	while ((++c)->datatype) {
		if (c->datatype && strncmp(datatype, c->datatype, 4))
			continue;
		break;
	}
	if (!c->type) goto notfound;

	/* fill in info and return */
	head->type = c->type;
	head->flags |= c->flags;
	return (0);

notfound:
	log_info("Type with '%.4s' data string was not implemented or not "
		"recognized.", datatype);
	head->type = 0;
	return (1);
}
